# Generated by Django 5.2 on 2025-01-25 13:29

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0082_alter_feature_dtype"),
    ]

    operations = [
        migrations.RunSQL(
            sql="""
                UPDATE lamindb_feature
                SET is_type = FALSE
                WHERE is_type IS NULL;

                UPDATE lamindb_flextable
                SET is_type = FALSE
                WHERE is_type IS NULL;

                UPDATE lamindb_param
                SET is_type = FALSE
                WHERE is_type IS NULL;

                UPDATE lamindb_project
                SET is_type = FALSE
                WHERE is_type IS NULL;

                UPDATE lamindb_reference
                SET is_type = FALSE
                WHERE is_type IS NULL;

                UPDATE lamindb_schema
                SET is_type = FALSE
                WHERE is_type IS NULL;

                UPDATE lamindb_ulabel
                SET is_type = FALSE
                WHERE is_type IS NULL;
            """
        ),
        migrations.AlterField(
            model_name="feature",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AlterField(
            model_name="flextable",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
    ]
