# Generated by Django 5.2 on 2025-01-21 17:03

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0080_polish_lamindbv1"),
    ]

    operations = [
        migrations.AlterField(
            model_name="collection",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
    ]
