# Generated by Django 5.2 on 2025-07-04 18:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0109_record_input_of_runs_alter_record_run_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="record",
            old_name="values_artifacts",
            new_name="linked_artifacts",
        ),
    ]
