# Generated by Django 5.2 on 2025-08-09 13:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0118_alter_recordproject_value_projectrecord"),
    ]

    operations = [
        migrations.RenameField(
            model_name="project",
            old_name="records",
            new_name="linked_in_records",
        ),
        migrations.AddField(
            model_name="project",
            name="records",
            field=models.ManyToManyField(
                related_name="projects",
                through="lamindb.ProjectRecord",
                to="lamindb.record",
            ),
        ),
    ]
