# Generated by Django 5.1.4 on 2025-04-30 09:11

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0090_runproject_project_runs"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="featurevalue",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="space",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="user",
            options={"base_manager_name": "objects"},
        ),
    ]
