# Generated by Django 5.1.7 on 2025-05-23 23:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0098_alter_feature_type_alter_project_type_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="writelog",
            name="seqno",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.RenameField(
            model_name="writelog",
            old_name="seqno",
            new_name="id",
        ),
    ]
