# Generated by Django 5.2 on 2025-05-11 18:54

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0096_remove_artifact__param_values_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="ArtifactParamValue",
        ),
        migrations.DeleteModel(
            name="SchemaParam",
        ),
        migrations.DeleteModel(
            name="Param",
        ),
        migrations.DeleteModel(
            name="ParamValue",
        ),
        migrations.DeleteModel(
            name="RunParamValue",
        ),
    ]
