# Generated by Django 5.2 on 2025-01-16 01:29

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0078_lamindbv1_part6c"),
    ]

    operations = [
        migrations.AlterField(
            model_name="rundata",
            name="value_json",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="tidytabledata",
            name="value_json",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="tidytable",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="_tidytables",
                to="lamindb.schema",
            ),
        ),
    ]
