# Generated by Django 5.2 on 2025-05-27 11:29

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0101_alter_artifact_hash_alter_feature_name_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="writelog",
            name="branch_code",
        ),
        migrations.RemoveField(
            model_name="writelog",
            name="space_uid",
        ),
        migrations.AddField(
            model_name="writelog",
            name="branch",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AddField(
            model_name="writelog",
            name="space",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="writelog",
            name="run_uid",
            field=models.CharField(default="0000000000000000", max_length=20),
        ),
        migrations.AlterField(
            model_name="writelog",
            name="record_uid",
            field=models.JSONField(db_index=True, default=0),
            preserve_default=False,
        ),
        migrations.AlterModelOptions(
            name="migrationstate",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="tablestate",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterField(
            model_name="writelog",
            name="migration_state",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="lamindb.migrationstate"
            ),
        ),
        migrations.AlterField(
            model_name="writelog",
            name="table",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="lamindb.tablestate"
            ),
        ),
    ]
