# Generated by Django 5.2 on 2025-01-27 13:48

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0084_alter_schemafeature_feature_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="feature",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="flextable",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, null=True
            ),
        ),
    ]
