# Generated by Django 5.2 on 2025-05-07 12:16

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0092_alter_artifactfeaturevalue_artifact_and_more"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="schemacomponent",
            unique_together={("composite", "slot"), ("composite", "slot", "component")},
        ),
    ]
